"""This module implements the ForEachBlockPass class."""
from __future__ import annotations

import copy
import logging
from typing import Any
from typing import Callable
from typing import Sequence

from distributed import get_client
from distributed import rejoin
from distributed import secede

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.machine import MachineModel
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.gates.constant.unitary import ConstantUnitaryGate
from bqskit.ir.gates.parameterized.pauli import PauliGate
from bqskit.ir.gates.parameterized.unitary import VariableUnitaryGate
from bqskit.ir.operation import Operation
from bqskit.ir.point import CircuitPoint
from bqskit.utils.typing import is_sequence

_logger = logging.getLogger(__name__)


class ForEachBlockPass(BasePass):
    """
    The ForEachBlockPass class.

    This is a control pass that executes another pass or passes on every block
    in the circuit. This will be done in parallel if executed in the compiler
    framework.
    """

    key = 'ForEachBlockPass_data'
    """The key in data, where block data will be put."""

    def __init__(
        self,
        loop_body: BasePass | Sequence[BasePass],
        calculate_error_bound: bool = False,
        collection_filter: Callable[[Operation], bool] | None = None,
        replace_filter: Callable[[Circuit, Operation], bool] | None = None,
    ) -> None:
        """
        Construct a ForEachBlockPass.

        Args:
            loop_body (BasePass | Sequence[BasePass]): The pass or passes
                to execute on every block.

            calculate_error_bound (bool): If set to true, will calculate
                errors on blocks after running `loop_body` on them and
                use these block errors to calculate an upper bound on the
                full circuit error. (Default: False)

            collection_filter (Callable[[Operation], bool] | None):
                A predicate that determines which operations should have
                `loop_body` called on them. Called with each operation
                in the circuit. If this returns true, that operation will
                be formed into an individual circuit and passed through
                `loop_body`. Defaults to all CircuitGates,
                ConstantUnitaryGates, and VariableUnitaryGates.

            replace_filter (Callable[[Circuit, Operation], bool] | None):
                A predicate that determines if the resulting circuit, after
                calling `loop_body` on a block, should replace the original
                operation. Called with the circuit output from `loop_body`
                and the original operation. If this returns true, the
                operation will be replaced with the new circuit.
                Defaults to always replace.

        Raises:
            ValueError: If a Sequence[BasePass] is given, but it is empty.
        """

        if not is_sequence(loop_body) and not isinstance(loop_body, BasePass):
            raise TypeError(
                'Expected Pass or sequence of Passes, got %s.'
                % type(loop_body),
            )

        if is_sequence(loop_body):
            truth_list = [isinstance(elem, BasePass) for elem in loop_body]
            if not all(truth_list):
                raise TypeError(
                    'Expected Pass or sequence of Passes, got %s.'
                    % type(loop_body[truth_list.index(False)]),
                )
            if len(loop_body) == 0:
                raise ValueError('Expected at least one pass.')

        self.loop_body = loop_body if is_sequence(loop_body) else [loop_body]
        self.calculate_error_bound = calculate_error_bound
        self.collection_filter = collection_filter or default_collection_filter
        self.replace_filter = replace_filter or default_replace_filter

        if not callable(self.collection_filter):
            raise TypeError(
                'Expected callable method that maps Operations to booleans for'
                ' collection_filter, got %s.' % type(self.collection_filter),
            )

        if not callable(self.replace_filter):
            raise TypeError(
                'Expected callable method that maps Circuit and Operations to'
                ' booleans for replace_filter'
                ', got %s.' % type(self.replace_filter),
            )

    def run(self, circuit: Circuit, data: dict[str, Any] = {}) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        # Make room in data for block data
        if self.key not in data:
            data[self.key] = []
        data[self.key].append([])

        # Collect blocks
        blocks: list[tuple[int, Operation]] = []
        for cycle, op in circuit.operations_with_cycles():
            if self.collection_filter(op):
                blocks.append((cycle, op))

        # Get the machine model
        model = BasePass.get_model(circuit, data)

        # Go through the blocks
        points: list[CircuitPoint] = []
        ops: list[Operation] = []
        subcircuits: list[Circuit] = []
        block_datas: list[dict[str, Any]] = []
        for i, (cycle, op) in enumerate(blocks):
            block_data: dict[str, Any] = {}

            # Form Subcircuit
            if isinstance(op.gate, CircuitGate):
                subcircuit = op.gate._circuit.copy()
                subcircuit.set_params(op.params)
            else:
                subcircuit = Circuit.from_operation(op)

            # Form Subtopology
            subnumbering = {op.location[i]: i for i in range(len(op.location))}
            subdata = {
                'machine_model': MachineModel(
                    len(op.location),
                    model.get_subgraph(op.location, subnumbering),
                ),
            }

            # Record Data Part 1
            block_data['op'] = copy.deepcopy(op)
            block_data['subcircuit_pre'] = subcircuit.copy()

            subcircuits.append(subcircuit)
            block_datas.append(block_data)

        # Perform Work
        if 'executor' in data:  # In Parallel
            for block_data in block_datas:
                block_data['executor'] = data['executor']
            client = get_client()
            completed_subcircuits = []
            completed_block_datas = []
            futures = []
            subc_futures = client.scatter(subcircuits)
            data_futures = client.scatter(block_datas)
            for subcircuit, block_data in zip(subc_futures, data_futures):
                future = client.submit(
                    _sub_do_work,
                    self.loop_body,
                    subcircuit,
                    block_data,
                )
                futures.append(future)
            secede()
            client.gather(futures)
            rejoin()
            for future in futures:
                x, y = future.result()
                completed_subcircuits.append(x)
                completed_block_datas.append(y)

        else:  # Sequentially
            completed_subcircuits = []
            completed_block_datas = []
            for subcircuit, block_data in zip(subcircuits, block_datas):
                x, y = _sub_do_work(self.loop_body, subcircuit, block_data)
                completed_subcircuits.append(x)
                completed_block_datas.append(y)

        # Process work
        for i, (cycle, op) in enumerate(blocks):
            subcircuit = completed_subcircuits[i]
            block_data = completed_block_datas[i]
            # Record Data Part 2
            block_data['subcircuit_post'] = subcircuit.copy()
            block_data['loop_body_data'] = subdata
            block_data['point'] = CircuitPoint(cycle, op.location[0])

            # Calculate Errors
            if self.calculate_error_bound:
                new_utry = subcircuit.get_unitary()
                old_utry = op.get_unitary()
                error = new_utry.get_distance_from(old_utry)
                block_data['error'] = error
                _logger.info(f'Block {i} has error {error}.')

            # Mark Blocks to be Replaced
            if self.replace_filter(subcircuit, op):
                _logger.info(f'Replacing block {i}.')
                points.append(CircuitPoint(cycle, op.location[0]))
                ops.append(
                    Operation(
                        CircuitGate(subcircuit, True),
                        op.location,
                        subcircuit.params,
                    ),
                )
                block_data['replaced'] = True
            else:
                block_data['replaced'] = False

            # Record block data into pass data
            data[self.key][-1].append(block_data)

        # Replace blocks
        circuit.batch_replace(points, ops)


def default_collection_filter(op: Operation) -> bool:
    return isinstance(
        op.gate, (
            CircuitGate,
            ConstantUnitaryGate,
            VariableUnitaryGate,
            PauliGate,
        ),
    )


def default_replace_filter(circuit: Circuit, op: Operation) -> bool:
    return True


def _sub_do_work(
    loop_body: Sequence[BasePass],
    subcircuit: Circuit,
    subdata: dict[str, Any],
) -> tuple[Circuit, dict[str, Any]]:
    for loop_pass in loop_body:
        loop_pass.run(subcircuit, subdata)
    return subcircuit, subdata



"""This module implements helper methods for random generators."""
from __future__ import annotations

import ctypes

import numpy as np


def seed_random_sources(seed: int) -> None:
    """Seed the various sources of randomness BQSKit uses."""
    # set rand() seed, used by Ceres
    libc = ctypes.CDLL(None)
    libc.srand(seed)
    # set numpy seed
    np.random.seed(seed)



"""This module implements the UnitaryMatrix class."""
from __future__ import annotations

import logging
from typing import Any
from typing import Sequence
from typing import Union

import numpy as np
import numpy.typing as npt
import scipy as sp
from scipy.stats import unitary_group

from bqskit.qis.state.state import StateLike
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.statemap import StateVectorMap
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitary import Unitary
from bqskit.utils.docs import building_docs
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_square_matrix
from bqskit.utils.typing import is_valid_radixes

if not building_docs():
    from numpy.lib.mixins import NDArrayOperatorsMixin
else:
    class NDArrayOperatorsMixin:  # type: ignore
        pass

_logger = logging.getLogger(__name__)


class UnitaryMatrix(Unitary, StateVectorMap, NDArrayOperatorsMixin):
    """A concrete representation of a unitary matrix."""

    def __init__(
        self,
        input: UnitaryLike,
        radixes: Sequence[int] = [],
        check_arguments: bool = True,
    ) -> None:
        """
        Constructs a `UnitaryMatrix` from the supplied unitary matrix.

        Args:
            input (UnitaryLike): The unitary matrix input.

            radixes (Sequence[int]): A sequence with its length equal to
                the number of qudits this `UnitaryMatrix` can act on. Each
                element specifies the base, number of orthogonal states,
                for the corresponding qudit. By default, the constructor
                will attempt to calculate `radixes` from `utry`.

            check_arguments (bool): If true, check arguments for type
                and value errors.

        Raises:
            ValueError: If `input` is not unitary.

            ValueError: If the dimension of `input` does not match the
                expected dimension from `radixes`.

            RuntimeError: If `radixes` is not specified and the
                constructor cannot infer it.

        Examples:
            >>> UnitaryMatrix(
            ...     [
            ...         [0, 1],
            ...         [1, 0],
            ...     ],
            ... )  # Creates a single-qubit Pauli X unitary matrix.
        """

        # Stop any actual logic when building documentation
        if building_docs():
            self._utry: npt.NDArray[np.complex128] = np.array([])
            return

        # Copy constructor
        if isinstance(input, UnitaryMatrix):
            self._utry = input.numpy
            self._radixes = input.radixes
            self._dim = input.dim
            return

        if check_arguments and not is_square_matrix(input):
            raise TypeError(f'Expected square matrix, got {type(input)}.')

        if check_arguments and not UnitaryMatrix.is_unitary(input):
            raise ValueError('Input failed unitary condition.')

        dim = len(input)

        if radixes:
            self._radixes = tuple(radixes)

        # Check if unitary dimension is a power of two
        elif dim & (dim - 1) == 0:
            self._radixes = tuple([2] * int(np.round(np.log2(dim))))

        # Check if unitary dimension is a power of three
        elif 3 ** int(np.round(np.log(dim) / np.log(3))) == dim:  # noqa
            radixes = [3] * int(np.round(np.log(dim) / np.log(3)))
            self._radixes = tuple(radixes)

        else:
            raise RuntimeError(
                'Unable to determine radixes'
                ' for UnitaryMatrix with dim %d.' % dim,
            )

        if check_arguments and not is_valid_radixes(self.radixes):
            raise TypeError('Invalid qudit radixes.')

        if check_arguments and np.prod(self.radixes) != dim:
            raise ValueError('Qudit radixes mismatch with dimension.')

        self._utry = np.array(input, dtype=np.complex128)
        self._dim = dim

    _num_params = 0

    @property
    def numpy(self) -> npt.NDArray[np.complex128]:
        """The NumPy array holding the unitary."""
        return self._utry

    @property
    def shape(self) -> tuple[int, ...]:
        """The two-dimensional square shape of the unitary."""
        return self._utry.shape

    @property
    def dtype(self) -> np.typing.DTypeLike:
        """The NumPy data type of the unitary."""
        return self._utry.dtype

    @property
    def T(self) -> UnitaryMatrix:
        """The transpose of the unitary."""
        return UnitaryMatrix(self._utry.T, self.radixes, False)

    @property
    def dagger(self) -> UnitaryMatrix:
        """The conjugate transpose of the unitary."""
        return self.conj().T

    def __len__(self) -> int:
        """The dimension of the square unitary matrix."""
        return self.shape[0]

    def __iter__(self) -> int:
        """An iterator that iterates through the rows of the matrix."""
        return self._utry.__iter__()

    def conj(self) -> UnitaryMatrix:
        """Return the complex conjugate unitary matrix."""
        return UnitaryMatrix(self._utry.conj(), self.radixes, False)

    def otimes(self, *utrys: UnitaryLike) -> UnitaryMatrix:
        """
        Calculate the tensor or kroneckor product with other unitaries.

        Args:
            utrys (UnitaryLike | Sequence[UnitaryLike]): The unitary or
                unitaries to computer the tensor product with.

        Returns:
            UnitaryMatrix: The resulting unitary matrix.
        """

        utrys = [UnitaryMatrix(u) for u in utrys]

        utry_acm = self.numpy
        radixes_acm = self.radixes
        for utry in utrys:
            utry_acm = np.kron(utry_acm, utry.numpy)
            radixes_acm += utry.radixes

        return UnitaryMatrix(utry_acm, radixes_acm)

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the same object, satisfies the :class:`Unitary` API."""
        return self

    def get_distance_from(self, other: UnitaryLike) -> float:
        """
        Return the distance between `self` and `other`.

        The distance is given as:

        .. math::

            \\sqrt{1 - \\frac{|Tr(U_1^\\dagger U_2)|}{N}^2}

        Args:
            other (UnitaryLike): The unitary to measure distance from.

        Returns:
            float: A value between 1 and 0, where 0 means the two unitaries
            are equal up to global phase and 1 means the two unitaries are
            very unsimilar or far apart.
        """
        other = UnitaryMatrix(other)
        num = np.abs(np.trace(self.conj().T @ other))
        dem = self.dim
        frac = min(num / dem, 1)
        dist = np.sqrt(1 - (frac ** 2))
        return dist if dist > 0.0 else 0.0

    def get_statevector(self, in_state: StateLike) -> StateVector:
        """
        Calculate the output state after applying this unitary to `in_state`.

        Args:
            in_state (StateLike): The input state to apply `self` to.

        Returns:
            StateVector: The output state.

        Raises:
            ValueError: If the state's dimension doesn't match the
                unitary's dimension.
        """
        if not StateVector.is_pure_state(in_state):
            raise TypeError(f'Expected StateVector, got {type(in_state)}.')

        in_state = StateVector(in_state)

        if in_state.dim != self.dim:
            raise ValueError(
                'State unitary dimension mismatch; '
                f'expected {self.dim}, got {in_state.dim}.',
            )

        vec = in_state[:, None]
        out_vec = self @ vec
        return StateVector(out_vec.reshape((-1,)), self.radixes)

    @staticmethod
    def identity(dim: int, radixes: Sequence[int] = []) -> UnitaryMatrix:
        """
        Construct an identity UnitaryMatrix.

        Args:
            dim (int): The dimension of the identity matrix.

            radixes (Sequence[int]): The number of orthogonal states
                for each qudit, defaults to qubits.

        Returns:
            UnitaryMatrix: An identity matrix.

        Raises:
            ValueError: If `dim` is nonpositive.
        """
        if dim <= 0:
            raise ValueError('Invalid dimension for identity matrix.')
        return UnitaryMatrix(np.identity(dim), radixes)

    @staticmethod
    def closest_to(
        M: npt.NDArray[np.complex128],
        radixes: Sequence[int] = [],
    ) -> UnitaryMatrix:
        """
        Calculate and return the closest unitary to a given matrix.

        Calculate the unitary matrix U that is closest with respect to the
        operator norm distance to the general matrix M.

        Args:
            M (np.ndarray): The matrix input.

            radixes (Sequence[int]): The radixes for the Unitary.

        Returns:
            (UnitaryMatrix): The unitary matrix closest to M.

        References:
            D.M.Reich. “Characterisation and Identification of Unitary Dynamics
            Maps in Terms of Their Action on Density Matrices”
        """
        if not is_square_matrix(M):
            raise TypeError('Expected square matrix.')

        V, _, Wh = sp.linalg.svd(M)
        return UnitaryMatrix(V @ Wh, radixes, False)

    @staticmethod
    def random(num_qudits: int, radixes: Sequence[int] = []) -> UnitaryMatrix:
        """
        Sample a random unitary from the haar distribution.

        Args:
            num_qudits (np.ndarray): The number of qudits for the matrix.
                This is not the dimension.

            radixes (Sequence[int]): The radixes for the Unitary.

        Returns:
            (UnitaryMatrix): A random unitary matrix.

        Raises:
            ValueError: If `num_qudits` is nonpositive.

            ValueError: If the length of `radixes` is not equal to
                `num_qudits`.
        """
        if not is_integer(num_qudits):
            raise TypeError(
                f'Expected int for num_qudits, got {type(num_qudits)}.',
            )

        if num_qudits <= 0:
            raise ValueError('Expected positive number for num_qudits.')

        radixes = tuple(radixes if len(radixes) > 0 else [2] * num_qudits)

        if not is_valid_radixes(radixes):
            raise TypeError('Invalid qudit radixes.')

        if len(radixes) != num_qudits:
            raise ValueError(
                'Expected length of radixes to be equal to num_qudits:'
                ' %d != %d' % (len(radixes), num_qudits),
            )

        U = unitary_group.rvs(int(np.prod(radixes)))
        return UnitaryMatrix(U, radixes, False)

    def __eq__(self, other: object) -> bool:
        """Check if `self` is approximately equal to `other`."""
        if isinstance(other, Unitary):
            other_unitary = other.get_unitary()
            if self.shape != other_unitary.shape:
                return False
            return np.allclose(self, other_unitary)

        if isinstance(other, np.ndarray):
            return np.allclose(self, other)

        return NotImplemented

    def save(self, filename: str) -> None:
        """Save the unitary to a file."""
        np.savetxt(filename, self.numpy)

    def __getitem__(
            self, index: Any,
    ) -> np.complex128 | npt.NDArray[np.complex128]:
        """Implements NumPy API for the StateVector class."""
        return self._utry[index]

    @staticmethod
    def from_file(filename: str) -> UnitaryMatrix:
        """Load a unitary from a file."""
        return UnitaryMatrix(np.loadtxt(filename, dtype=np.complex128))

    @staticmethod
    def is_unitary(U: np.typing.ArrayLike, tol: float = 1e-8) -> bool:
        """
        Check if U is a unitary matrix.

        Args:
            U (np.typing.ArrayLike): The matrix to check.

            tol (float): The numerical precision of the check.

        Returns:
            bool: True if U is a unitary.
        """

        if isinstance(U, UnitaryMatrix):
            return True

        if not isinstance(U, np.ndarray):
            U = np.array(U)

        if not is_square_matrix(U):
            return False

        X = U @ U.conj().T
        Y = U.conj().T @ U
        I = np.identity(X.shape[0])

        if not np.allclose(X, I, rtol=0, atol=tol):
            if _logger.isEnabledFor(logging.DEBUG):
                norm = np.linalg.norm(X - I)
                _logger.debug(
                    'Failed unitary condition, ||UU^d - I|| = %e' %
                    norm,
                )
            return False

        if not np.allclose(Y, I, rtol=0, atol=tol):
            if _logger.isEnabledFor(logging.DEBUG):
                norm = np.linalg.norm(Y - I)
                _logger.debug(
                    'Failed unitary condition, ||U^dU - I|| = %e' %
                    norm,
                )
            return False

        return True

    def __array__(
        self,
        dtype: np.typing.DTypeLike = np.complex128,
    ) -> npt.NDArray[np.complex128]:
        """Implements NumPy API for the UnitaryMatrix class."""
        if dtype != np.complex128:
            raise ValueError('UnitaryMatrix only supports Complex128 dtype.')

        return self._utry

    def __array_ufunc__(
        self,
        ufunc: np.ufunc,
        method: str,
        *inputs: npt.NDArray[Any],
        **kwargs: Any,
    ) -> UnitaryMatrix | npt.NDArray[np.complex128]:
        """Implements NumPy API for the UnitaryMatrix class."""
        if method != '__call__':
            return NotImplemented

        non_unitary_involved = False
        args: list[npt.NDArray[Any]] = []
        for input in inputs:
            if isinstance(input, UnitaryMatrix):
                args.append(input.numpy)
            else:
                args.append(input)
                non_unitary_involved = True

        out = ufunc(*args, **kwargs)

        # The results are unitary
        # if only unitaries are involved
        # and unitaries are closed under the specific operation.
        convert_back = (
            not non_unitary_involved and (
                ufunc.__name__ == 'conjugate'
                or ufunc.__name__ == 'matmul'
                or ufunc.__name__ == 'negative'
                or ufunc.__name__ == 'positive'
            )
            or (
                ufunc.__name__ == 'multiply'
                and all(
                    np.isscalar(input) or isinstance(input, UnitaryMatrix)
                    for input in inputs
                )
                and all(
                    np.abs(np.abs(input) - 1) <= 1e-14
                    for input in inputs if np.isscalar(input)
                )
            )
        )

        if convert_back:
            return UnitaryMatrix(out, self.radixes)

        return out

    def __str__(self) -> str:
        """Return the string representation of the unitary."""
        return str(self._utry)

    def __repr__(self) -> str:
        """Return the repr representation of the unitary."""
        return repr(self._utry)

    def __hash__(self) -> int:
        """Return the hash of the unitary."""
        return hash((self._utry[0][0], self._utry[-1][-1], self.shape))


UnitaryLike = Union[UnitaryMatrix, np.ndarray, Sequence[Sequence[Any]]]



"""This module defines the PassPredicate base class."""
from __future__ import annotations

import abc
from typing import Any

from bqskit.ir.circuit import Circuit


class PassPredicate(abc.ABC):
    """
    The PassPredicate abstract base class.

    A PassPredicate implements the :func:`get_truth_value` method, which is
    called from control passes to determine the flow of execution.
    """

    @abc.abstractmethod
    def get_truth_value(self, circuit: Circuit, data: dict[str, Any]) -> bool:
        """Call this predicate and retrieve the truth value result."""

    def __call__(self, circuit: Circuit, data: dict[str, Any]) -> bool:
        """Call this predicate and retrieve the truth value result."""

        if not isinstance(circuit, Circuit):
            raise TypeError('Expected Circuit, got %s.' % type(circuit))

        if not isinstance(data, dict):
            raise TypeError('Expected dictionary, got %s.' % type(data))

        return self.get_truth_value(circuit, data)



"""This module defines the QuickPartitioner pass."""
from __future__ import annotations

import heapq
import logging
from typing import Any

from bqskit.compiler.basepass import BasePass
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.region import CircuitRegion
from bqskit.utils.typing import is_integer

_logger = logging.getLogger(__name__)


class QuickPartitioner(BasePass):
    """
    The QuickPartitioner Pass.

    This pass forms partitions in the circuit by iterating over the operations
    in a topological order and binning them into blocks.
    """

    def __init__(
        self,
        block_size: int = 3,
    ) -> None:
        """
        Construct a QuickPartitioner.

        Args:
            block_size (int): Maximum size of partitioned blocks.
                (Default: 3)

        Raises:
            ValueError: If `block_size` is less than 2.
        """

        if not is_integer(block_size):
            raise TypeError(
                f'Expected integer for block_size, got {type(block_size)}.',
            )

        if block_size < 2:
            raise ValueError(
                f'Expected block_size to be greater than 2, got {block_size}.',
            )

        self.block_size = block_size

    def run(self, circuit: Circuit, data: dict[str, Any] = {}) -> None:
        """
        Partition gates in a circuit into a series of CircuitGates.

        Args:
            circuit (Circuit): Circuit to be partitioned.

            data (dict[str,Any]): Optional data unique to specific run.
        """

        # Number of qudits in the circuit
        num_qudits = circuit.num_qudits

        # If block size > circuit size, return the circuit as a block
        if self.block_size > num_qudits:
            _logger.warning(
                'Configured block size is greater than circuit size; '
                'blocking entire circuit.',
            )
            circuit.fold({
                qudit_index: (0, circuit.num_cycles)
                for qudit_index in range(circuit.num_qudits)
            })
            return

        # List to hold the active blocks
        active_blocks: Any = []

        # List to hold the finished blocks
        finished_blocks: Any = {}
        block_id = 0

        # Active qudit cycles and block-qudit dependencies
        qudit_actives: Any = [{} for _ in range(num_qudits)]
        qudit_dependencies: Any = [{} for _ in range(num_qudits)]

        # The partitioned circuit
        partitioned_circuit = Circuit(num_qudits, circuit.radixes)

        # For each cycle, operation in topological order
        for cycle, op in circuit.operations_with_cycles():

            # Get the qudits of the operation
            qudits = op.location._location

            # Update the active locations of the qudits
            for qudit in qudits:
                qudit_actives[qudit][cycle] = None

            # Compile a list of admissible blocks out of the
            # active blocks for the operation
            admissible_blocks = []
            for index, block in enumerate(active_blocks):
                if all([qudit not in block[-1] for qudit in qudits]):
                    admissible_blocks.append(index)

            # Boolean indicator to capture if an active block
            # has been found for the operation
            found = False

            # For all admissible blocks, check if all operation
            # qudits are in the block. If such a block is found,
            # update the upper region bound for the corresponding
            # qudits, and raise the found boolean
            for block in [active_blocks[index] for index in admissible_blocks]:
                if all([qudit in block for qudit in qudits]):
                    for qudit in qudits:
                        block[qudit][1] = cycle
                    found = True
                    break

            updated_qudits = set()

            # If such a block is not found
            if not found:

                # For all admissible blocks, check if the operation
                # qudits can be added to the block without breaching
                # the size limit. If such a block is found, add the
                # new qudits, update the region bounds, check if any
                # blocks are finished, and raise the found boolean
                for block in [
                    active_blocks[index]
                    for index in admissible_blocks
                ]:
                    if len(set(list(qudits) + list(block.keys()))) - \
                            1 <= self.block_size:
                        for qudit in qudits:
                            if qudit not in block:
                                block[qudit] = [cycle, cycle]
                            else:
                                block[qudit][1] = cycle
                        block_id, updated_qudits = self.compute_finished_blocks(
                            block, qudits, active_blocks,
                            finished_blocks, block_id, qudit_dependencies, cycle, num_qudits,  # noqa
                        )
                        found = True
                        break

            # If a block is still not found, check if any blocks are finished
            # with the new operation qudits, create a new block, and add it
            # to the list of active blocks
            if not found:
                block_id, updated_qudits = self.compute_finished_blocks(
                    None, qudits, active_blocks,
                    finished_blocks, block_id, qudit_dependencies, cycle, num_qudits,  # noqa
                )
                block = {qudit: [cycle, cycle] for qudit in qudits}
                block[-1] = set()
                active_blocks.append(block)

            # Where the active qudit cycles keep getting updated
            while updated_qudits:

                # Check if any blocks corresponding to updated qudits
                # are eligible to be added to the circuit. If eligible,
                # update actives, dependencies, and updated qudits.
                final_regions = []
                new_updated_qudits = set()
                for qudit in updated_qudits:
                    blk_ids = list(qudit_dependencies[qudit].keys())
                    for blk_id in blk_ids:
                        num_passed = 0
                        for qdt, bounds in finished_blocks[blk_id].items():
                            if len(qudit_actives[qdt]) == 0:
                                num_passed += 1
                            elif next(iter(qudit_actives[qdt])) == bounds[0]:
                                num_passed += 1
                        if num_passed == len(finished_blocks[blk_id]):
                            for qdt, bounds in finished_blocks[blk_id].items():
                                for cycle in range(bounds[0], bounds[1] + 1):
                                    if cycle in qudit_actives[qdt]:
                                        del qudit_actives[qdt][cycle]
                                del qudit_dependencies[qdt][blk_id]
                                new_updated_qudits.add(qdt)
                            final_regions.append(
                                CircuitRegion(
                                {qdt: (bounds[0], bounds[1]) for qdt, bounds in finished_blocks[blk_id].items()},  # noqa
                                ),
                            )
                            del finished_blocks[blk_id]

                # If there are any regions
                if final_regions:

                    # Sort the regions if multiple exist
                    if len(final_regions) > 1:
                        final_regions = self.topo_sort(final_regions)

                    # Fold the final regions into a partitioned circuit
                    for region in final_regions:
                        region = circuit.downsize_region(region)
                        cgc = circuit.get_slice(region.points)
                        partitioned_circuit.append_gate(
                            CircuitGate(
                                cgc, True,
                            ), sorted(
                                list(
                                    region.keys(),
                                ),
                            ), list(
                                cgc.params,
                            ),
                        )

                updated_qudits = new_updated_qudits

        # Convert all remaining finished blocks and active blocks
        # into circuit regions
        final_regions = []
        for block in finished_blocks.values():
            final_regions.append(
                CircuitRegion(
                {qdt: (bounds[0], bounds[1]) for qdt, bounds in block.items()},  # noqa
                ),
            )
        for block in active_blocks:
            del block[-1]
            final_regions.append(
                CircuitRegion(
                {qdt: (bounds[0], bounds[1]) for qdt, bounds in block.items()},  # noqa
                ),
            )

        # If there are any regions
        if final_regions:

            # Sort the regions if multiple exist
            if len(final_regions) > 1:
                final_regions = self.topo_sort(final_regions)

            # Fold the final regions into a partitioned circuit
            for region in final_regions:
                region = circuit.downsize_region(region)
                cgc = circuit.get_slice(region.points)
                partitioned_circuit.append_gate(
                    CircuitGate(
                        cgc, True,
                    ), sorted(
                        list(
                            region.keys(),
                        ),
                    ), list(
                        cgc.params,
                    ),
                )

        # Copy the partitioned circuit to the original circuit
        circuit.become(partitioned_circuit)

    def compute_finished_blocks(  # type: ignore
        self, block, qudits, active_blocks, finished_blocks,
        block_id, qudit_dependencies, cycle, num_qudits,
    ):
        """Add blocks with all inactive qudits to the finished_blocks list and
        remove them from the active_blocks list."""

        # Compile the qudits from the new operation,
        # the active qudits of the block being updated,
        # and the qudits in the block's inadmissible list
        qudits = set(qudits)
        if block:
            qudits.update([qudit for qudit in block if qudit != -1])
            qudits.update(block[-1])

        remove_blocks = []

        # For all active blocks
        for active_block in active_blocks:

            # If the active block is different than the block being updated
            if active_block != block:

                # If any of the qudits are in the active block or its
                # inadmissible list, then add those qudits to the
                # inadmissible list of the active block
                if any([
                    qudit in active_block or qudit in active_block[-1]
                    for qudit in qudits
                ]):
                    active_block[-1].update(qudits)

                # If the active block has reached its maximum size
                # and/or all of its qudits are inadmissible,
                # then add it to the remove list
                if (
                    len(active_block) - 1 == self.block_size and  # noqa
                    all([
                        qudit in active_block[-1]
                        for qudit in active_block if qudit != -1
                    ])
                ) or (
                    cycle - max(
                        active_block[qudit][1]
                        for qudit in active_block
                        if qudit != -1
                    ) > 200
                ) or len(active_block[-1]) == num_qudits:
                    remove_blocks.append(active_block)

        # Remove all blocks in the remove list from the active list
        # and add them to the finished blocks list after deleting
        # their inadmissible list and update qudit dependencies
        updated_qudits = set()
        for remove_block in remove_blocks:
            del remove_block[-1]
            finished_blocks[block_id] = remove_block
            for qudit in remove_block:
                qudit_dependencies[qudit][block_id] = None
                updated_qudits.add(qudit)
            active_blocks.remove(remove_block)
            block_id += 1

        return block_id, updated_qudits

    def topo_sort(self, regions):  # type: ignore
        """Topologically sort circuit regions."""

        # Number of regions in the circuit
        num_regions = len(regions)

        # For each region, generate the number of in edges
        # and the list of all out edges
        in_edges = [0] * num_regions
        out_edges: Any = [[] for _ in range(num_regions)]
        for i in range(num_regions - 1):
            for j in range(i + 1, num_regions):
                dependency = regions[i].dependency(regions[j])
                if dependency == 1:
                    in_edges[i] += 1
                    out_edges[j].append(i)
                elif dependency == -1:
                    in_edges[j] += 1
                    out_edges[i].append(j)

        # Convert the list of number of in edges in to a min-heap
        in_edges = [[num_edges, i] for i, num_edges in enumerate(in_edges)]
        heapq.heapify(in_edges)

        index = 0
        sorted_regions = []

        # While there are regions remaining to be sorted
        while index < num_regions:

            # Select the regions with zero remaining in edges
            selections = []
            while in_edges and not in_edges[0][0]:
                selections.append(heapq.heappop(in_edges))

            if not selections:
                raise RuntimeError('Unable to topologically sort regions.')

            # Add the regions to the sorted list
            for region in selections:
                sorted_regions.append(regions[region[1]])
                index += 1

            # Remove the regions from all other regions' in edges counts
            for i in range(len(in_edges)):
                in_edges[i][0] -= sum(
                    in_edges[i][1] in out_edges[region[1]]
                    for region in selections
                )

            # Convert in edges into a min-heap
            heapq.heapify(in_edges)

        return sorted_regions



"""
=======================================================
Quantum Information Science Objects (:mod:`bqskit.qis`)
=======================================================

.. currentmodule:: bqskit.qis

The `bqskit.qis` package contains class definitions
for common objects found in quantum information science.

The most widely used will be the `UnitaryMatrix` object which
represents a pure quantum operation in matrix form.

.. rubric:: Abstract Base Classes

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:
    :template: autosummary/class_no_init.rst

    Unitary
    DifferentiableUnitary
    LocallyOptimizableUnitary
    StateVectorMap

.. rubric:: Core Classes

.. autosummary::
    :toctree: autogen
    :recursive:

    UnitaryMatrix
    UnitaryBuilder
    StateVector
    PauliMatrices
    PermutationMatrix

.. rubric:: Type Aliases

.. autosummary::
    :toctree: autogen
    :recursive:

    UnitaryLike
    StateLike
    RealVector
"""
from __future__ import annotations

from bqskit.qis.pauli import PauliMatrices
from bqskit.qis.permutation import PermutationMatrix
from bqskit.qis.state import StateLike
from bqskit.qis.state import StateVector
from bqskit.qis.state import StateVectorMap
from bqskit.qis.unitary import DifferentiableUnitary
from bqskit.qis.unitary import LocallyOptimizableUnitary
from bqskit.qis.unitary import RealVector
from bqskit.qis.unitary import Unitary
from bqskit.qis.unitary import UnitaryBuilder
from bqskit.qis.unitary import UnitaryLike
from bqskit.qis.unitary import UnitaryMatrix

__all__ = [
    'Unitary',
    'DifferentiableUnitary',
    'LocallyOptimizableUnitary',
    'UnitaryBuilder',
    'UnitaryMatrix',
    'StateVector',
    'UnitaryLike',
    'StateVectorMap',
    'PauliMatrices',
    'StateLike',
    'PermutationMatrix',
    'RealVector',
]



"""This module implements the GateCountPredicate class."""
from __future__ import annotations

import logging
from typing import Any

from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.passes.control.predicate import PassPredicate

_logger = logging.getLogger(__name__)


class GateCountPredicate(PassPredicate):
    """
    The GateCountPredicate class.

    The GateCountPredicate returns true if the number of one type of gate has
    changed in the circuit.
    """

    key = 'GateCountPredicate_circuit_count'

    def __init__(self, gate: Gate) -> None:
        """Construct a GateCountPredicate."""

        if not isinstance(gate, Gate):
            raise TypeError(f'Expected gate, got {type(gate)}')

        self.gate = gate

    def get_truth_value(self, circuit: Circuit, data: dict[str, Any]) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        gate_count = circuit.count(self.gate)

        # If first call, record data and return true
        if self.key not in data:
            _logger.debug(f'Could not find {self.key} key.')
            data[self.key] = gate_count
            return True

        # Otherwise, check for a change and return accordingly
        if data[self.key] == gate_count:
            _logger.debug('Counts match; no change detected.')
            return False

        data[self.key] = gate_count
        _logger.debug('Counts do not match; change detected.')
        return True



"""This module implements the SynthesisPass abstract class."""
from __future__ import annotations

import logging
from abc import abstractmethod
from typing import Any
from typing import Callable

from dask.distributed import as_completed
from dask.distributed import Client
from dask.distributed import Future
from dask.distributed import rejoin
from dask.distributed import secede

from bqskit.compiler.basepass import BasePass
from bqskit.ir.circuit import Circuit
from bqskit.qis.state.state import StateLike
from bqskit.qis.unitary.unitarymatrix import UnitaryLike
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix

_logger = logging.getLogger(__name__)


class SynthesisPass(BasePass):
    """
    SynthesisPass class.

    The SynthesisPass is a base class that exposes an abstract
    synthesize function. Inherit from this class and implement the
    synthesize function to create a synthesis tool.

    A SynthesisPass will synthesize a new circuit targeting the input
    circuit's unitary.
    """

    @abstractmethod
    def synthesize(self, utry: UnitaryMatrix, data: dict[str, Any]) -> Circuit:
        """
        Synthesis abstract method to synthesize a UnitaryMatrix into a Circuit.

        Args:
            utry (UnitaryMatrix): The unitary to synthesize.

            data (Dict[str, Any]): Associated data for the pass.
                Can be used to provide auxillary information from
                previous passes. This function should never error based
                on what is in this dictionary.

        Note:
            This function should be self-contained and have no side effects.
        """

    def run(self, circuit: Circuit, data: dict[str, Any] = {}) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        target_utry = circuit.get_unitary()
        circuit.become(self.synthesize(target_utry, data))

    def batched_instantiate(
        self,
        circuits: list[Circuit],
        target: UnitaryLike | StateLike,
        client: Client,
        **kwargs: Any,
    ) -> list[list[Future]]:
        """
        Batch instantiate `circuits`.

        Args:
            circuits (list[Circuit]): The circuit batch to instantiate.

            target (UnitaryLike | StateLike): The instantiation target.

            client (Client): The Dask client used to submit jobs.

            kwargs (Any): Other keyword arguments are passed directly to
                instantiate calls.

        Returns:
            (list[list[Future]]): The Dask Futures corresponding to the
                the instantiate jobs submitted. This returns a double list,
                where the other list is indexed by input circuits, and the
                inner list is indexed by multistart.
        """
        multistarts = 1
        if 'multistarts' in kwargs:
            multistarts = kwargs['multistarts']
            kwargs['multistarts'] = 1

        futures: list[list[Future]] = []

        client.scatter(circuits)

        for circuit in circuits:
            futures.append([])
            for i in range(multistarts):
                futures[-1].append(
                    client.submit(
                        Circuit.instantiate,
                        circuit,
                        pure=False,
                        target=target,
                        **kwargs,
                    ),
                )

        return futures

    def gather_best_results(
        self,
        futures: list[list[Future]],
        client: Client,
        fn: Callable[..., float],
        *args: Any,
        **kwargs: Any,
    ) -> list[Circuit]:
        """
        Gather best results from a `batched_instantiate` call.

        Args:
            futures (list[list[Future]]): The futures return from a
                `batched_instantiate` call.

            client (Client): The Dask client used to submit the jobs.

            fn (Callable[..., float]): The function used to sort
                instantiated circuits. This should take a circuit as
                the first parameter.

            args (Any): Arguments passed directly to fn.

            kwargs (Any): Keyword arguments passed directly to fn.

        Returns:
            (list[Circuit]): The resulting circuits. There is one circuit
                for each future list, i.e., `len(output) == len(futures)`.
        """

        score_futures: list[list[Future]] = []
        for future_list in futures:
            score_futures.append([])
            for future in future_list:
                score_futures[-1].append(
                    client.submit(
                        fn,
                        future,
                        *args,
                        **kwargs,
                    ),
                )

        flat_score_list = []
        for future_list in score_futures:
            flat_score_list.extend(future_list)

        secede()
        client.gather(flat_score_list)
        rejoin()

        best_circuit_futures = []
        for i, score_list in enumerate(score_futures):
            scores = client.gather(score_list)
            best_index = scores.index(min(scores))
            best_circuit_futures.append(futures[i][best_index])

        return client.gather(best_circuit_futures)

    def gather_first_results(
        self,
        futures: list[list[Future]],
        client: Client,
    ) -> list[Circuit]:
        """
        Gather first results from a `batched_instantiate` call.

        Args:
            futures (list[list[Future]]): The futures return from a
                `batched_instantiate` call.

            client (Client): The Dask client used to submit the jobs.

        Returns:
            (list[Circuit]): The resulting circuits. There is one circuit
                for each future list, i.e., `len(output) == len(futures)`.
                This call will return the first multistart instantiate
                job for each circuit.
        """
        all_futures = []
        future_to_index_map = {}
        for index, future_list in enumerate(futures):
            all_futures.extend(future_list)
            for future in future_list:
                future_to_index_map[future] = index

        secede()
        circuits = []
        indices_seen = set()
        for future, circuit in as_completed(all_futures, with_results=True):
            index = future_to_index_map[future]
            if index not in indices_seen:
                indices_seen.add(index)
                circuits.append((index, circuit))
                for other_future in futures[index]:
                    if not other_future.done():
                        client.cancel(other_future)
        rejoin()

        return [c for _, c in sorted(circuits, key=lambda x: x[0])]



# type: ignore
"""This module implements a OPENQASM parse tree visitor."""
from __future__ import annotations

import logging
from typing import Any
from typing import NamedTuple

import lark
import numpy as np
from lark import Visitor

from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.ir.gates.composed.controlled import ControlledGate
from bqskit.ir.gates.composed.daggergate import DaggerGate
from bqskit.ir.gates.constant.ch import CHGate
from bqskit.ir.gates.constant.cx import CXGate
from bqskit.ir.gates.constant.cy import CYGate
from bqskit.ir.gates.constant.cz import CZGate
from bqskit.ir.gates.constant.h import HGate
from bqskit.ir.gates.constant.identity import IdentityGate
from bqskit.ir.gates.constant.s import SGate
from bqskit.ir.gates.constant.sdg import SdgGate
from bqskit.ir.gates.constant.swap import SwapGate
from bqskit.ir.gates.constant.sx import SXGate
from bqskit.ir.gates.constant.t import TGate
from bqskit.ir.gates.constant.tdg import TdgGate
from bqskit.ir.gates.constant.x import XGate
from bqskit.ir.gates.constant.y import YGate
from bqskit.ir.gates.constant.z import ZGate
from bqskit.ir.gates.parameterized.rx import RXGate
from bqskit.ir.gates.parameterized.rxx import RXXGate
from bqskit.ir.gates.parameterized.ry import RYGate
from bqskit.ir.gates.parameterized.ryy import RYYGate
from bqskit.ir.gates.parameterized.rz import RZGate
from bqskit.ir.gates.parameterized.rzz import RZZGate
from bqskit.ir.gates.parameterized.u1 import U1Gate
from bqskit.ir.gates.parameterized.u2 import U2Gate
from bqskit.ir.gates.parameterized.u3 import U3Gate
from bqskit.ir.lang.language import LangException
from bqskit.ir.operation import Operation
from bqskit.utils.typing import is_iterable


_logger = logging.getLogger(__name__)


class QubitReg(NamedTuple):
    """Definition of a Qubit Register."""
    name: str
    size: int


class GateDef(NamedTuple):
    """Definition of a QASM Gate."""
    qasm_name: str
    num_params: int
    num_vars: int
    gate: Gate


class OPENQASMVisitor(Visitor):
    """
    The OPENQASMVisitor class.

    This class defines a complete visitor for a parse tree produced by
    OPENQASMParser.
    """

    def __init__(self) -> None:
        self.op_list: list[Operation] = []
        self.qubit_regs: list[QubitReg] = []
        self.gate_defs: dict[str, GateDef] = {}
        self.fill_gate_defs()
        self.gate_def_parsing_obj = None

    def get_circuit(self) -> Circuit:
        """Retrieve the circuit generated after walking a parse tree."""
        num_qubits = sum(qubit_reg.size for qubit_reg in self.qubit_regs)
        if num_qubits == 0:
            raise LangException('No qubit registers defined.')
        circuit = Circuit(num_qubits)
        circuit.extend(self.op_list)
        return circuit

    def fill_gate_defs(self) -> None:
        """Prefills gate definitions with built-in gates."""
        self.gate_defs['U'] = GateDef('U', 3, 1, U3Gate())
        self.gate_defs['u'] = GateDef('u', 3, 1, U3Gate())
        self.gate_defs['u3'] = GateDef('u3', 3, 1, U3Gate())
        self.gate_defs['u2'] = GateDef('u2', 2, 1, U2Gate())
        self.gate_defs['u1'] = GateDef('u1', 1, 1, U1Gate())
        self.gate_defs['cx'] = GateDef('cx', 0, 2, CXGate())
        self.gate_defs['cy'] = GateDef('cy', 0, 2, CYGate())
        self.gate_defs['cz'] = GateDef('cz', 0, 2, CZGate())
        self.gate_defs['ch'] = GateDef('ch', 0, 2, CHGate())
        self.gate_defs['swap'] = GateDef('swap', 0, 2, SwapGate())
        self.gate_defs['id'] = GateDef('id', 0, 1, IdentityGate(1))
        self.gate_defs['x'] = GateDef('x', 0, 1, XGate())
        self.gate_defs['y'] = GateDef('y', 0, 1, YGate())
        self.gate_defs['z'] = GateDef('z', 0, 1, ZGate())
        self.gate_defs['h'] = GateDef('h', 0, 1, HGate())
        self.gate_defs['s'] = GateDef('s', 0, 1, SGate())
        self.gate_defs['sdg'] = GateDef('sdg', 0, 1, SdgGate())
        self.gate_defs['t'] = GateDef('t', 0, 1, TGate())
        self.gate_defs['tdg'] = GateDef('tdg', 0, 1, TdgGate())
        self.gate_defs['rx'] = GateDef('rx', 1, 1, RXGate())
        self.gate_defs['ry'] = GateDef('ry', 1, 1, RYGate())
        self.gate_defs['rz'] = GateDef('rz', 1, 1, RZGate())
        self.gate_defs['sx'] = GateDef('sx', 0, 1, SXGate())
        self.gate_defs['sxdg'] = GateDef('sxdg', 0, 1, DaggerGate(SXGate()))
        self.gate_defs['rxx'] = GateDef('rxx', 1, 2, RXXGate())
        self.gate_defs['ryy'] = GateDef('ryy', 1, 2, RYYGate())
        self.gate_defs['rzz'] = GateDef('rzz', 1, 2, RZZGate())
        self.gate_defs['cu1'] = GateDef('cu1', 1, 2, ControlledGate(U1Gate()))
        self.gate_defs['cu2'] = GateDef('cu2', 2, 2, ControlledGate(U2Gate()))
        self.gate_defs['cu3'] = GateDef('cu3', 3, 2, ControlledGate(U3Gate()))

    def qreg(self, args: Any) -> None:
        reg_name = args.children[0]
        if any(reg_name == reg.name for reg in self.qubit_regs):
            raise LangException('Qubit register redeclared: %s.' % reg_name)
        reg_size = int(args.children[1])
        reg = QubitReg(reg_name, reg_size)
        _logger.debug('Qubit register %s declared with size %d.' % reg)
        self.qubit_regs.append(reg)

    def gate(self, args: Any) -> None:
        gate_name = args.children[0]
        param_list = None if len(args.children) == 2 else args.children[1]
        if len(args.children) == 2:
            var_list = args.children[1]
        else:
            var_list = args.children[2]

        if gate_name not in self.gate_defs:
            raise LangException('Unrecognized gate: %s.' % gate_name)

        gate_def = self.gate_defs[gate_name]

        # Calculate params
        params = []
        if param_list is not None:
            if is_iterable(eval_explist(param_list)):
                params = [float(i) for i in eval_explist(param_list)]
            else:
                params = [float(eval_explist(param_list))]

        if len(params) != gate_def.num_params:
            raise LangException(
                'Expected %d params got %d params for gate %s.'
                % (gate_def.num_params, len(params), gate_name),
            )

        # Calculate location
        location = []
        for reg_name, reg_idx in qubits_from_list(var_list):
            outer_idx = 0
            for reg in self.qubit_regs:
                if reg.name == reg_name:
                    location.append(outer_idx + reg_idx)
                    break
                outer_idx += reg.num_qudits

        if len(location) != gate_def.num_vars:
            raise LangException(
                'Gate acts on %d qubits, got %d qubit variables.'
                % (gate_def.num_vars, len(location)),
            )

        # Calculate operation
        self.op_list.append(Operation(gate_def.gate, location, params))

    def cxgate(self, args: Any) -> None:
        self.gate(args)

    def ugate(self, args: Any) -> None:
        self.gate(args)

    def gatedecl(self, args: Any) -> None:
        _logger.warning('Ignoring unsupported qasm feature: gate declaration.')

    def creg(self, args: Any) -> None:
        _logger.warning('Ignoring unsupported qasm feature: classic register.')

    def measure(self, args: Any) -> None:
        _logger.warning('Ignoring unsupported qasm feature: measure.')

    def reset(self, args: Any) -> None:
        _logger.warning('Ignoring unsupported qasm feature: reset.')


eval_locals = {
    'pi': np.pi,
    'sin': np.sin,
    'cos': np.cos,
    'tan': np.tan,
    'ln': np.log,
    'exp': np.exp,
}


def eval_exp_recurse(tree: lark.Tree) -> Any:
    if isinstance(tree, lark.Token):
        return str(tree)
    code = ''
    for op in tree.children:
        if isinstance(op, lark.Token):
            code += str(op)
            continue
        if op.data == 'unaryexp':
            unaryop = op.children[0]
            code += f'{unaryop.children[0]}({eval_exp_recurse(op.children[1])})'
            continue
        elif op.data == 'usub':
            code += '-'
        elif op.data == 'pow':
            base = eval_exp_recurse(op.children[0])
            exp = eval_exp_recurse(op.children[1])
            code += f'{base}**{exp}'
            continue
        code += ' '.join(map(eval_exp_recurse, op.children))
    return code


def eval_exp(tree: lark.Tree) -> Any:
    return eval(eval_exp_recurse(tree), {}, eval_locals)


def eval_explist(tree: lark.Tree) -> Any:
    if isinstance(tree, list):
        tree = tree[0]
    if tree.data == 'explist':
        if len(tree.children) == 1:
            return eval_exp(tree.children[0])
        list_vals = eval_explist(tree.children[0])
        exp_vals = eval_exp(tree.children[1])
        if isinstance(list_vals, tuple):
            return (*list_vals, exp_vals)
        return list_vals, exp_vals
    elif tree.data == 'exp':
        return eval_exp(tree)
    else:
        raise Exception()


def qubits_from_list(tree: lark.Tree) -> list[tuple[str, int]]:
    if tree.data == 'anylist':
        return qubits_from_list(tree.children[0])
    elif tree.data == 'mixedlist':
        q2 = (str(tree.children[-2]), int(tree.children[-1]))
        if isinstance(tree.children[0], lark.Tree):
            q1 = qubits_from_list(tree.children[0])
            return [*q1, q2]
        return [q2]
    else:
        raise Exception()



"""This package implements predicates for use in control passes."""
from __future__ import annotations

from bqskit.passes.control.predicates.change import ChangePredicate
from bqskit.passes.control.predicates.count import GateCountPredicate
from bqskit.passes.control.predicates.notpredicate import NotPredicate

__all__ = ['ChangePredicate', 'GateCountPredicate', 'NotPredicate']



"""
BQSKit Gates (:mod:`bqskit.ir.gates`)
=====================================

.. rubric:: Gate Base Classes

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:
    :template: autosummary/gate.rst

    ComposedGate
    QubitGate
    QutritGate
    ConstantGate

.. rubric:: Constant Gates

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:
    :template: autosummary/gate.rst

    CHGate
    CPIGate
    CSUMGate
    CNOTGate
    CXGate
    CYGate
    CZGate
    HGate
    IdentityGate
    SGate
    SdgGate
    SqrtCNOTGate
    SwapGate
    SqrtXGate
    SXGate
    TGate
    TdgGate
    ConstantUnitaryGate
    XGate
    XXGate
    YGate
    YYGate
    ZGate
    ZZGate

.. rubric:: Parameterized Gates

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:
    :template: autosummary/gate.rst

    PauliGate
    RXGate
    RXXGate
    RYGate
    RYYGate
    RZGate
    RZZGate
    U1Gate
    U2Gate
    U3Gate
    U8Gate
    VariableUnitaryGate

.. rubric:: Composed Gates

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:
    :template: autosummary/gate.rst

    ControlledGate
    DaggerGate
    FrozenParameterGate
    TaggedGate
    VariableLocationGate

.. rubric:: Special Gates

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:

    CircuitGate
"""
from __future__ import annotations

from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.gates.composed import *  # noqa
from bqskit.ir.gates.composed import __all__ as composed_all
from bqskit.ir.gates.composedgate import ComposedGate
from bqskit.ir.gates.constant import *  # noqa
from bqskit.ir.gates.constant import __all__ as constant_all
from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.parameterized import *  # noqa
from bqskit.ir.gates.parameterized import __all__ as parameterized_all
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.ir.gates.qutritgate import QutritGate

__all__ = composed_all + constant_all + parameterized_all
__all__ += ['ComposedGate', 'QubitGate', 'QutritGate', 'ConstantGate']
__all__ += ['CircuitGate']



"""This module implements the Instantiater base class."""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from bqskit.qis.state.state import StateVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit
    from bqskit.qis.state.state import StateLike
    from bqskit.qis.unitary.unitarymatrix import UnitaryLike


class Instantiater(abc.ABC):
    """
    The Instantiater class.

    An Instantiater is responsible for instantiating circuit templates such that
    the resulting circuit bests implements the desired target.
    """

    @abc.abstractmethod
    def instantiate(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector,
        x0: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float64]:
        """
        Instantiate `circuit` to best implement `target`.

        Args:
            circuit (Circuit): The circuit template to instantiate.

            target (UnitaryMatrix | StateVector): The unitary matrix to
                implement or state to prepare.

            x0 (np.ndarray): Initial point to use during instantiation.

        Returns:
            (np.ndarray): The paremeters for the circuit that makes the
                circuit best implement `target`.

        Notes:
            This method should be side-effect free. This is necessary since
            many instantiate calls to the same circuit using the same
            Instantiater object may happen in parallel.
        """

    @staticmethod
    @abc.abstractmethod
    def is_capable(circuit: Circuit) -> bool:
        """Return true if the circuit can be instantiated."""

    @staticmethod
    @abc.abstractmethod
    def get_violation_report(circuit: Circuit) -> str:
        """
        Return a message explaining why `circuit` cannot be instantiated.

        Args:
            circuit (Circuit): Generate a report for this circuit.

        Raises:
            ValueError: If `circuit` can be instantiated with this
                instantiater.
        """

    def gen_starting_points(
        self,
        multistarts: int,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector,
    ) -> list[npt.NDArray[np.float64]]:
        """
        Generate `multistarts` starting points for instantiation.

        Args:
            multistarts (int): The number of starting points to generate.

            circuit (Circuit): The circuit to generate the points for.

            target (UnitaryMatrix | StateVector): The target.

        Return:
            (list[np.ndarray]): List of starting inputs for instantiation.

        Raises:
            ValueError: If `multistarts` is not a positive integer.
        """
        if not is_integer(multistarts):
            raise TypeError(
                'Expected int for multistarts, got %s.' % type(multistarts),
            )

        if multistarts <= 0:
            raise ValueError(
                'Expected positive integer for multistarts'
                ', got %d' % multistarts,
            )

        return [  # [ circuit.params ] + [  # TODO: re-evaluate
            np.random.random(circuit.num_params)
            for i in range(multistarts)
        ]

    def check_target(
        self,
        target: UnitaryLike | StateLike,
    ) -> UnitaryMatrix | StateVector:
        """Check `target` to be valid and return it casted."""
        try:
            typed_target = StateVector(target)  # type: ignore
        except (ValueError, TypeError):
            try:
                typed_target = UnitaryMatrix(target)  # type: ignore
            except (ValueError, TypeError) as ex:
                raise TypeError(
                    'Expected either StateVector, UnitaryMatrix, or'
                    ' CostFunction for target, got %s.' % type(target),
                ) from ex

        return typed_target

    @staticmethod
    @abc.abstractmethod
    def get_method_name() -> str:
        """Return the name of this method."""



"""This module implements the DoWhileLoopPass class."""
from __future__ import annotations

import logging
from typing import Any
from typing import Sequence

from bqskit.compiler.basepass import BasePass
from bqskit.ir.circuit import Circuit
from bqskit.passes.control.predicate import PassPredicate
from bqskit.utils.typing import is_sequence

_logger = logging.getLogger(__name__)


class DoWhileLoopPass(BasePass):
    """
    The DoWhileLoopPass class.

    This is a control pass that executes a pass and then conditionally executes
    it again in a loop.
    """

    def __init__(
        self,
        condition: PassPredicate,
        loop_body: BasePass | Sequence[BasePass],
    ) -> None:
        """
        Construct a DoWhileLoopPass.

        Args:
            condition (PassPredicate): The condition checked.

            loop_body (BasePass | Sequence[BasePass]): The pass or passes
                to execute while `condition` is true.

        Raises:
            ValueError: If a Sequence[BasePass] is given, but it is empty.
        """

        if not isinstance(condition, PassPredicate):
            raise TypeError('Expected PassPredicate, got %s.' % type(condition))

        if not is_sequence(loop_body) and not isinstance(loop_body, BasePass):
            raise TypeError(
                'Expected Pass or sequence of Passes, got %s.'
                % type(loop_body),
            )

        if is_sequence(loop_body):
            truth_list = [isinstance(elem, BasePass) for elem in loop_body]
            if not all(truth_list):
                raise TypeError(
                    'Expected Pass or sequence of Passes, got %s.'
                    % type(loop_body[truth_list.index(False)]),
                )
            if len(loop_body) == 0:
                raise ValueError('Expected at least one pass.')

        self.condition = condition
        self.loop_body = loop_body if is_sequence(loop_body) else [loop_body]

    def run(self, circuit: Circuit, data: dict[str, Any] = {}) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        # Perform Work
        _logger.debug('Loop body executing...')
        for loop_pass in self.loop_body:
            loop_pass.run(circuit, data)

        while self.condition(circuit, data):
            _logger.debug('Loop body executing...')
            for loop_pass in self.loop_body:
                loop_pass.run(circuit, data)



"""This module implements the WindowOptimizationPass."""
from __future__ import annotations

import logging
from typing import Any
from typing import Callable

import numpy as np

from bqskit.compiler.basepass import BasePass
from bqskit.ir.circuit import Circuit
from bqskit.passes.synthesis.qsearch import QSearchSynthesisPass
from bqskit.passes.synthesis.synthesis import SynthesisPass
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_iterable

_logger = logging.getLogger(__name__)


class WindowOptimizationPass(BasePass):
    """
    The WindowOptimizationPass class.

    Previously run passes can leave window markers. This pass will select a
    window around that marker and resynthesize, potentially for a depth
    reduction.
    """

    def __init__(
        self,
        window_size: int = 9,
        synthesispass: SynthesisPass = QSearchSynthesisPass(),
        replace_filter: Callable[[Circuit, Circuit], bool] | None = None,
    ) -> None:
        """
        Construct a WindowOptimizationPass.

        Args:
            window_size (int): The size of the window to surround each
                marker. (Default: 9)

            synthesispass (SynthesisPass): The configured synthesis algorithm
                to use during window synthesis.

            replace_filter (Callable[[Circuit, Circuit], bool] | None):
                A predicate that determines if the synthesis result should
                replace the original operation. The first parameter is the
                output from synthesis and the second parameter is the original
                window. If this returns true, the newly synthesized circuit
                will replace the original circuit. Defaults to always replace.

        Raises:
            ValueError: If `window_size` is <= 1.
        """

        if not is_integer(window_size):
            raise TypeError(
                'Expected integer for success_threshold'
                ', got %s' % type(window_size),
            )

        if window_size <= 1:
            raise ValueError(
                'Expected integer greater than 1 for window_size'
                ', got %d.' % window_size,
            )

        if not isinstance(synthesispass, SynthesisPass):
            raise TypeError(
                'Expected SynthesisPass, got %s.' % type(synthesispass),
            )

        self.replace_filter = replace_filter or default_replace_filter

        if not callable(self.replace_filter):
            raise TypeError(
                'Expected callable method for replace_filter'
                ', got %s.' % type(self.replace_filter),
            )

        self.window_size = window_size
        self.synthesispass = synthesispass

    def run(self, circuit: Circuit, data: dict[str, Any] = {}) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        # Check data for windows markers
        if 'window_markers' not in data:
            _logger.warning('Did not find any window markers.')
            return

        window_markers = data['window_markers']
        _logger.debug('Found window_markers: %s.' % str(window_markers))

        if not is_iterable(window_markers):
            _logger.debug('Invalid type for window_markers.')
            return

        if not all(is_integer(marker) for marker in window_markers):
            _logger.debug('Invalid type for window_markers.')
            return

        # Resynthesis each window
        index_shift = 0
        for marker in window_markers:
            marker -= index_shift

            # Slice window
            begin_cycle = int(marker - self.window_size // 2)
            end_cycle = int(marker + np.ceil(self.window_size / 2))

            if begin_cycle < 0:
                begin_cycle = 0

            if end_cycle > circuit.num_cycles:
                end_cycle = circuit.num_cycles - 1

            window = Circuit(circuit.num_qudits, circuit.radixes)
            window.extend(circuit[begin_cycle:end_cycle])

            _logger.info(
                'Resynthesizing window from cycle '
                f'{begin_cycle} to {end_cycle}.',
            )

            # Synthesize
            utry = window.get_unitary()
            new_window = self.synthesispass.synthesize(utry, data)

            # Replace
            if self.replace_filter(new_window, window):
                _logger.debug('Replacing window with synthesized circuit.')

                actual_window_size = end_cycle - begin_cycle
                for _ in range(actual_window_size):
                    circuit.pop_cycle(begin_cycle)

                circuit.insert_circuit(
                    begin_cycle,
                    new_window,
                    list(range(circuit.num_qudits)),
                )

                index_shift += actual_window_size - new_window.num_cycles


def default_replace_filter(new_circuit: Circuit, og_window: Circuit) -> bool:
    return True



"""This test will ensure that the circuit's built by QFAST can be minimized
successfully."""
from __future__ import annotations

from scipy.stats import unitary_group

from bqskit.ir.circuit import Circuit
from bqskit.ir.gates import PauliGate
from bqskit.ir.opt.cost.functions import HilbertSchmidtResiduals
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


def test_full_pauli_gate() -> None:
    circuit = Circuit(3)
    circuit.append_gate(PauliGate(3), [0, 1, 2])
    cost = HilbertSchmidtResiduals(circuit, UnitaryMatrix(unitary_group.rvs(8)))
    circuit.minimize(cost)
    assert cost.get_cost(circuit.params) < 1e-6



"""This module tests the math library."""
from __future__ import annotations

from typing import Any

import numpy as np
import numpy.typing as npt
import pytest
import scipy as sp
from scipy.stats import unitary_group

from bqskit.qis.pauli import PauliMatrices
from bqskit.utils.math import dexpmv
from bqskit.utils.math import dot_product
from bqskit.utils.math import pauli_expansion
from bqskit.utils.math import softmax
from bqskit.utils.math import unitary_log_no_i


def dexpm_exact(
    M: npt.NDArray[np.complex128],
    dM: npt.NDArray[np.complex128],
    term_count: int = 100,
) -> tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:
    """
    Exact matrix exponential derivative calculation.

    Rossmann 2002 Theorem 5 Section 1.2
    """
    adjMp = dM
    total = np.zeros(M.shape, dtype=np.complex128)

    for i in range(term_count):
        total += adjMp
        adjMp = (M @ adjMp) - (adjMp @ M)
        adjMp *= -1
        adjMp /= i + 2

    F = sp.linalg.expm(M)
    dF = F @ total
    return F, dF


class TestDexpmv:

    @pytest.mark.parametrize(
        'alpha', [np.random.random(16) for i in range(100)],
    )
    def test_single(self, alpha: npt.NDArray[np.float64]) -> None:
        paulis = PauliMatrices(2)
        H = paulis.dot_product(alpha)

        for p in paulis:
            F0, dF0 = dexpm_exact(H, p)
            F1, dF1 = dexpmv(H, p)

            assert np.allclose(F0, F1)
            assert np.allclose(dF0, dF1)

    @pytest.mark.parametrize(
        'alpha', [np.random.random(16) for i in range(100)],
    )
    def test_vector(self, alpha: npt.NDArray[np.float64]) -> None:
        paulis = PauliMatrices(2)
        H = paulis.dot_product(alpha)

        dFs0 = []
        for p in paulis:
            _, dF = dexpm_exact(H, p)
            dFs0.append(dF)

        dFs0_np = np.array(dFs0)

        _, dFs1 = dexpmv(H, paulis.numpy)

        assert np.allclose(dFs0_np, dFs1)

    def test_invalid(self) -> None:
        with pytest.raises(Exception):
            dexpmv(0, 0)  # type: ignore

        with pytest.raises(Exception):
            dexpmv(0, [1, 0])  # type: ignore

        with pytest.raises(Exception):
            dexpmv([1, 0], 0)  # type: ignore

        with pytest.raises(Exception):
            dexpmv([1, 0], [1, 0])  # type: ignore

        I = np.identity(2)

        with pytest.raises(Exception):
            dexpmv(I, 0)  # type: ignore

        with pytest.raises(Exception):
            dexpmv(0, I)  # type: ignore

        with pytest.raises(Exception):
            dexpmv(I, [1, 0])  # type: ignore

        with pytest.raises(Exception):
            dexpmv([1, 0], I)  # type: ignore


class TestSoftmax:

    @pytest.mark.parametrize('x', [np.random.random(100) for i in range(100)])
    def test_1(self, x: npt.NDArray[np.float64]) -> None:
        assert np.abs(np.sum(softmax(10 * x)) - 1) < 1e-15

    def test_2(self) -> None:
        x = np.ones(10)
        x[0] = 2
        assert np.argmax(softmax(x)) == 0
        x[0] = 1
        x[5] = 2
        assert np.argmax(softmax(x)) == 5

    @pytest.mark.parametrize('test_variable', ['a', False, True])
    def test_invalid(self, test_variable: Any) -> None:
        with pytest.raises(TypeError):
            softmax(test_variable)


class TestDotProduct:
    def test_valid_1(self) -> None:
        sigma = np.array([[[0, 1], [1, 0]], [[1, 0], [0, 1]]])
        alpha = np.array([1, 1])

        expected = np.array([[1, 1], [1, 1]])
        assert np.allclose(dot_product(alpha, sigma), expected)

    def test_valid_2(self) -> None:
        sigma = np.array([[[0, 1], [1, 0]], [[1, 0], [0, 1]]])
        alpha = np.array([0.5, 0.5])

        expected = 0.5 * np.array([[1, 1], [1, 1]])
        assert np.allclose(dot_product(alpha, sigma), expected)

    def test_invalid(self) -> None:
        sigma = 'a'
        alpha = 'b'

        with pytest.raises(TypeError):
            dot_product(alpha, sigma)  # type: ignore


class TestUnitaryLog:

    @pytest.mark.parametrize(
        'reU',
        PauliMatrices(1).paulis
        + PauliMatrices(2).paulis
        + PauliMatrices(3).paulis
        + PauliMatrices(4).paulis
        + [unitary_group.rvs(16) for _ in range(100)],
    )
    def test_valid(self, reU: npt.NDArray[np.complex128]) -> None:
        H = unitary_log_no_i(reU)
        assert np.allclose(H, H.conj().T, rtol=0, atol=1e-15)
        U = sp.linalg.expm(1j * H)
        assert 1 - (np.abs(np.trace(U.conj().T @ reU)) / U.shape[0]) <= 1e-15
        assert np.allclose(
            U.conj().T @ U,
            np.identity(len(U)),
            rtol=0,
            atol=1e-14,
        )
        assert np.allclose(
            U @ U.conj().T,
            np.identity(len(U)),
            rtol=0,
            atol=1e-14,
        )


class TestPauliExpansion:

    @pytest.mark.parametrize(
        'reH',
        PauliMatrices(1).paulis
        + PauliMatrices(2).paulis
        + PauliMatrices(3).paulis
        + PauliMatrices(4).paulis,
    )
    def test_valid(self, reH: npt.NDArray[np.complex128]) -> None:
        alpha = pauli_expansion(reH)
        print(alpha)
        H = PauliMatrices(int(np.log2(reH.shape[0]))).dot_product(alpha)
        assert np.linalg.norm(H - reH) < 1e-16
